import configparser
from pathlib import Path

import matplotlib.pyplot as plt
import sklearn

from geocif import geocif as gc
from geocif import logger as log

plt.style.use("default")
sklearn.set_config(transform_output="pandas")

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


def run(inputs, logger, parser, section, item, type, values):
    if type == "str":
        original_value = parser.get(section, item)
    elif type == "bool":
        original_value = parser.getboolean(section, item)
    elif type == "int":
        original_value = parser.getint(section, item)
    elif type == "float":
        original_value = parser.getfloat(section, item)

    for value in values:
        if type == "str":
            parser.set(section, item, value)
        elif type == "bool":
            parser.set(section, item, str(value))
        elif type == "int":
            parser.set(section, item, str(value))
        elif type == "float":
            parser.set(section, item, str(value))

        gc.execute_models(inputs, logger, parser)

    parser.set(section, item, original_value)

    return parser


def main(path_config_files=[Path("../config/geocif.txt")]):
    logger, parser = log.setup_logger_parser(path_config_files)
    inputs = gc.gather_inputs(parser)

    logger.info("=============================")
    logger.info("\tStarting GEOCIF Experiments")
    logger.info("=============================")

    # Experiment 1: include_lat_lon
    logger.info("Experiment 1: include_lat_lon")
    parser = run(
        inputs, logger, parser, "experiments", "include_lat_lon", "bool", [True, False]
    )

    # Experiment 2: feature_selection
    logger.info("Experiment 2: feature_selection")
    parser = run(
        inputs,
        logger,
        parser,
        "experiments",
        "feature_selection",
        "str",
        ["SelectKBest", "BorutaPy", "Leshy", "RFECV", "RFE"],
    )

    # Experiment 3: lag_years
    logger.info("Experiment 3: lag_years")
    parser = run(
        inputs, logger, parser, "experiments", "lag_years", "int", [1, 2, 3, 4, 5]
    )

    # Experiment 4: lag_yield_as_feature
    logger.info("Experiment 4: lag_yield_as_feature")
    parser = run(
        inputs,
        logger,
        parser,
        "experiments",
        "lag_yield_as_feature",
        "bool",
        [True, False],
    )

    # Experiment 5: median_years
    logger.info("Experiment 5: median_years")
    parser = run(
        inputs, logger, parser, "experiments", "median_years", "int", [2, 3, 4, 5]
    )

    # Experiment 6: median_yield_as_feature
    logger.info("Experiment 6: median_yield_as_feature")
    parser = run(
        inputs,
        logger,
        parser,
        "experiments",
        "median_yield_as_feature",
        "bool",
        [True, False],
    )

    # Experiment 7: analogous_year_yield_as_feature
    logger.info("Experiment 7: analogous_year_yield_as_feature")
    parser = run(
        inputs,
        logger,
        parser,
        "experiments",
        "analogous_year_yield_as_feature",
        "bool",
        [True, False],
    )

    # Experiment 8: optimize
    logger.info("Experiment 8: optimize")
    parser = run(
        inputs, logger, parser, "experiments", "optimize", "bool", [True, False]
    )


if __name__ == "__main__":
    main()



import numpy as np
from tqdm import tqdm
from sklearn.ensemble import RandomForestRegressor


def select_features(X, y, method="RFE", min_features_to_select=3):
    """

    Args:
        X:
        y:
        method:
        min_features_to_select:
        num_cpus:

    Returns:

    """

    # df = X.copy()
    #
    # # Initialize and apply StandardScaler
    # scaler = StandardScaler()
    # scaled_data = scaler.fit_transform(df)
    #
    # # Initialize and apply VarianceThreshold
    # # Note: Since data is standardized, all features now have variance of 1 before applying VarianceThreshold.
    # # You would adjust the threshold based on new criteria since variances have been normalized.
    # selector = VarianceThreshold(threshold=scaled_data.var().mean())
    # X = selector.fit_transform(scaled_data)
    selector = None

    # Fill in columns with median of that column
    X = X.fillna(X.median())

    # Define the RandomForestRegressor
    forest = RandomForestRegressor(
        n_estimators=500,
        n_jobs=8,
        max_depth=5,
        random_state=1,
    )

    # Adjusting numpy types due to deprecation warnings or errors
    np.int = np.int32
    np.float = np.float64
    np.bool = np.bool_

    if method == "SelectKBest":
        from sklearn.feature_selection import SelectKBest, f_regression

        k = 15  # Number of features to select
        selector = SelectKBest(score_func=f_regression, k=k)

        # Fit the selector to the data and transform the data to select the best features
        try:
            X_new = selector.fit_transform(X, y)
        except:
            breakpoint()

        # Get the selected feature indices
        selected_features = selector.get_support(indices=True)

        # Get the selected feature names
        selected_features = X.columns[selected_features].tolist()
    elif method == "SHAP":
        import pandas as pd
        from catboost import CatBoostRegressor
        from fasttreeshap import TreeExplainer as FastTreeExplainer
        from sklearn.model_selection import cross_val_score

        model = CatBoostRegressor(n_estimators=500, verbose=0, use_best_model=False)
        model.fit(X, y)

        explainer = FastTreeExplainer(model)
        shap_values = explainer.shap_values(X)

        # Step 5: Summarize the SHAP values for feature importance
        shap_importances = np.mean(np.abs(shap_values), axis=0)
        shap_importance_df = pd.DataFrame(
            {"feature": X.columns, "importance": shap_importances}
        ).sort_values(by="importance", ascending=False)

        def evaluate_model_with_n_features(N, X_train, y_train):
            top_features = shap_importance_df["feature"].head(N).values
            X_train_selected = X_train[top_features]
            selector = CatBoostRegressor(n_estimators=500, random_state=42, verbose=0)
            scores = cross_val_score(
                selector,
                X_train_selected,
                y_train,
                cv=5,
                scoring="neg_mean_squared_error",
                n_jobs=-1,
            )

            return np.mean(scores)

        # Evaluate model performance with different number of features
        nrange = [5, 10, 15, 20, 25, 30]
        cv_scores = []
        for N in tqdm(nrange):
            cv_scores.append(evaluate_model_with_n_features(N, X, y))

        # Select the number of features that gives the best cross-validation score (lowest MSE)
        optimal_N = nrange[np.argmax(cv_scores)]

        # Use optimal N to select features
        selected_features = (
            shap_importance_df["feature"].head(optimal_N).values.tolist()
        )
    elif method == "feature_engine":
        from feature_engine.selection import SmartCorrelatedSelection

        selector = SmartCorrelatedSelection(
            method="pearson",
            threshold=0.7,
            selection_method="model_performance",
            estimator=forest,
            scoring="neg_mean_squared_error",
        )

        X_filtered = selector.fit_transform(X, y)
        selected_features = X_filtered.columns.tolist()
    elif method == "RFECV":
        from sklearn.feature_selection import RFECV
        from sklearn.model_selection import KFold

        # Initialize a k-fold cross-validation strategy
        cv = KFold(n_splits=5)

        # Patch the scoring function to add a progress bar
        class RFECVWithProgress(RFECV):
            def _fit(self, X, y):
                from tqdm import tqdm

                n_features = X.shape[1]
                with tqdm(total=n_features) as pbar:

                    def patched_scorer(*args, **kwargs):
                        pbar.update(1)
                        return self.scorer_(*args, **kwargs)

                    self.scorer_ = patched_scorer
                    super()._fit(X, y)

        # Initialize RFECV with the estimator and cross-validation strategy
        selector = RFECVWithProgress(
            estimator=forest, step=1, n_jobs=-1, cv=cv, scoring="neg_mean_squared_error"
        )
        selector.fit(X, y)
        # Get the selected feature indices
        selected_features = selector.get_support(indices=True)

        # Get the selected feature names
        selected_features = X.columns[selected_features].tolist()
    elif method == "BorutaPy":
        from boruta import BorutaPy

        selector = BorutaPy(forest, n_estimators="auto", random_state=42)
        selector.fit(X.values, y.values)
        selected_features_mask = selector.support_
        selected_features = X.columns[selected_features_mask].tolist()
        tentative_features = X.columns[selector.support_weak_].tolist()

        selected_features = selected_features + tentative_features
    elif method == "Leshy":
        import arfs.feature_selection.allrelevant as arfsgroot
        from catboost import CatBoostRegressor

        model = CatBoostRegressor(n_estimators=350, verbose=0, use_best_model=False)
        selector = arfsgroot.Leshy(
            model,
            n_estimators="auto",
            verbose=1,
            max_iter=10,
            random_state=42,
            importance="fastshap",
        )
        selector.fit(X, y, sample_weight=None)

        selected_features = selector.get_feature_names_out()
        # feat_selector.plot_importance(n_feat_per_inch=5)
    elif method == "PowerShap":
        from powershap import PowerShap
        from catboost import CatBoostRegressor

        selector = PowerShap(
            model=CatBoostRegressor(n_estimators=500, verbose=0, use_best_model=True),
            power_alpha=0.05,
        )

        selector.fit(X, y)  # Fit the PowerShap feature selector
        selector.transform(X)  # Reduce the dataset to the selected features
    elif method == "BorutaShap":
        from BorutaShap import BorutaShap
        from catboost import CatBoostRegressor

        hyperparams = {
            "depth": 6,
            "learning_rate": 0.05,
            "iterations": 500,
            "subsample": 1.0,
            "random_strength": 0.5,
            "reg_lambda": 0.001,
            "loss_function": "RMSE",
            "early_stopping_rounds": 25,
            "random_seed": 42,
            "verbose": False,
        }
        model = CatBoostRegressor(**hyperparams)

        selector = BorutaShap(
            model=model, importance_measure="shap", classification=False
        )
        selector.fit(
            X=X,
            y=y,
            n_trials=100,
            sample=False,
            train_or_test="test",
            normalize=True,
            verbose=False,
        )
        selected_features_mask = selector.Subset().columns
        selected_features = X[selected_features_mask].columns.tolist()
    elif method == "Genetic":
        from sklearn_genetic import GAFeatureSelectionCV

        selector = GAFeatureSelectionCV(
            estimator=forest,
            cv=5,
            verbose=1,
            scoring="neg_mean_squared_error",
            max_features=max(len(X.columns) // 3, min_features_to_select),
            population_size=100,
            generations=40,
            crossover_probability=0.9,
            mutation_probability=0.1,
            keep_top_k=2,
            elitism=True,
            n_jobs=-1,
        )
        selector.fit(X, y)
        selected_features_mask = selector.support_
        selected_features = X.columns[selected_features_mask].tolist()
    elif method == "RFE":
        from sklearn.feature_selection import RFE

        selector = RFE(
            forest, n_features_to_select=min_features_to_select, step=1, verbose=1
        )
        selector = selector.fit(X, y)
        selected_features_mask = selector.support_
        selected_features = X.columns[selected_features_mask].tolist()
    else:
        raise ValueError("Method not recognized. Use BorutaPy, Genetic, or RFE")
    # tentative_features = X.columns[selector.support_weak_].tolist()

    # Filter the dataset for selected features
    X_filtered = X.loc[:, selected_features]

    return selector, X_filtered, selected_features



import pickle
import sqlite3

import pandas as pd

from geocif import utils


def config_to_dict(parser):
    """
    Reads a configuration file and returns the configuration as a nested dictionary.

    :param config_file_path: Path to the configuration file.
    :return: Dictionary with section names as keys and dictionaries of options and values as values.
    """
    # Initialize an empty dictionary to store the configuration
    config_dict = {}

    # Iterate over all sections and options, storing them in the dictionary
    for section in parser.sections():
        # Initialize the section dictionary
        section_dict = {}

        # Iterate over options in the current section
        for option in parser.options(section):
            # Get the value for the current option
            value = parser.get(section, option)
            # Store the option and its value in the section dictionary
            section_dict[option] = value

        # Store the current section in the main configuration dictionary
        config_dict[section] = section_dict

    return config_dict


def pprint_config(dict_config):
    """
    Pretty print the configuration file
    """
    import pprint

    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(dict_config)


def config_to_db(db_path, parser, today):
    """
    Store the configuration file in the database
    Args:
        db_path:
        parser:
        today:

    Returns:

    """
    con = sqlite3.connect(db_path)

    # Prepare a list to store section, option, and value
    data = []

    # Iterate through each section and each option within the section
    for section in parser.sections():
        for option in parser.options(section):
            # Append a tuple of section, option, and value to the list
            data.append((section, option, parser.get(section, option)))

    # Convert the list of tuples into a DataFrame
    df_parser = pd.DataFrame(data, columns=["Section", "Option", "Value"])
    df_parser.loc[:, "Now"] = today

    # name the index level
    df_parser.index.set_names(["Index"], inplace=True)
    utils.to_db(db_path, f"config_{today}", df_parser)

    con.commit()
    con.close()


def store(db_path, experiment_id, df, model, model_name):
    """

    Args:
        db_path:
        experiment_id:
        df:
        model:
        model_name:

    Returns:

    """
    con = sqlite3.connect(db_path)

    # Convert any categorical columns to the values else we get the error
    # Exception: 'Categorical' with dtype category does not support reduction 'all'
    for col in df.columns:
        if isinstance(df[col], pd.CategoricalDtype):
            df[col] = df[col].astype(str)

    # Change all categorical columns to type object
    for col in df.select_dtypes(include=["category"]).columns:
        df[col] = df[col].astype(str)

    # Output results to database
    try:
        utils.to_db(db_path, experiment_id, df)
    except Exception as e:
        print(f"Error: {e}")

    index_columns = ["Country", "Region", "Crop", "Harvest Year", "Stages"]
    # Output model pickle as a blob to database
    df_model = pd.DataFrame(
        {
            "Experiment_ID": [experiment_id],
            "Model": [model_name],
            "Model_Blob": [pickle.dumps(model)],
        }
    )
    # df_model.index = df_model.apply(
    #     lambda row: "_".join([str(row[col]) for col in index_columns]), axis=1
    # )

    # name the index level
    try:
        df_model.index.set_names(["Index"], inplace=True)
        utils.to_db(db_path, "models", df_model)
    except Exception as e:
        print(f"Error: {e}")

    con.commit()
    con.close()



import pandas as pd
from tqdm import tqdm

from geocif.backup import geo
from geocif.agmet import plot, utils


def create_title_for_plot(obj):
    """

    Args:
        obj:

    Returns:

    """
    region_name = obj.region.replace("_", " ").title()
    calendar_region_name = obj.calendar_region.replace("_", " ").title()
    country_name = obj.country.replace("_", " ").title()
    long_crop_name = utils.get_crop_name(obj.crop)

    # Get name of crop based on metadata/crop_per_season.csv file
    df_crop_per_season = pd.read_csv(obj.dir_metadata / "crop_per_season.csv")

    crop_name = df_crop_per_season[
        (df_crop_per_season["country"] == obj.country)
        & (df_crop_per_season["crop"] == obj.crop)
        & (df_crop_per_season["season"] == int(obj.growing_season))
    ]["name"].values

    crop_name = crop_name[0] if crop_name else long_crop_name.replace("_", " ").title()

    title_line_1 = f"{region_name} ({calendar_region_name}, {country_name})"
    title_line_2 = f"{crop_name} {obj.plot_season}"

    sup_title = f"{title_line_1}\n{title_line_2}"

    return sup_title


def loop_agmet(path_config_file=None):
    """
    Args:
    """
    # Create geo object
    obj = geo.geo(path_config_file)

    # Read in data on crop names in each country
    obj.read_statistics(read_countries=True)

    # Create combinations of run parameters
    all_combinations = obj.create_run_combinations()

    pbar = tqdm(all_combinations, total=len(all_combinations))
    for country, scale, crop, growing_season in pbar:
        # Setup country information
        obj.setup_country(country, scale, crop, growing_season)

        # Loop through seasons and plot for each admin_1 region and calendar region
        for plot_season in obj.plot_seasons:
            # Get list of the years that are closest to the year which we want to plot
            obj.get_closest_season(plot_season)

            for region in obj.list_regions:
                # Setup the region information
                obj.setup_region(region, plot_season, "region")

                # If available, add CHIRPS-GEFS data to the dataframe
                if obj.precip_var == "chirps":
                    obj.add_precip_forecast(plot_season)

                # Get crop calendar dates for region
                dates_calendar = [
                    obj.date_planting,
                    obj.date_greenup,
                    obj.date_senescence,
                    obj.date_harvesting,
                ]

                # TODO: Only plot if dates_calendar has valid dates
                # Create title for plot
                sup_title = create_title_for_plot(obj)

                ###############################################################
                # Agmet plots for regions
                ###############################################################
                # D:\Dropbox\Projects\GEOGLAM\crop_condition\AMIS\argentina
                plot.plots_ts_cur_yr(
                    obj.df_region,
                    obj.eo_plot,
                    closest=obj.closest,
                    dates_cal=dates_calendar,
                    frcast_yr=obj.plot_season,
                    logos=[obj.logo_harvest, obj.logo_geoglam],
                    dir_out=obj.dir_agmet / obj.scale,
                    sup_title=sup_title,
                    fname=f"{obj.region}.png",
                )

                ###############################################################
                # Agmet plots for calendar regions
                ###############################################################
                columns = obj.eo_model + ["month", "day", "yield"]
                if "chirps" in obj.df_region.columns:
                    bool_year_check, bool_date_check = obj.check_date(
                        obj.df_region, obj.plot_season
                    )

                    if bool_date_check and bool_year_check:
                        columns = obj.eo_model + [
                            "month",
                            "day",
                            "chirps_gefs",
                            "yield",
                        ]

                df_calendar_region = (
                    obj.df_region.groupby(
                        [
                            "country",
                            "calendar_region",
                            "harvest_season",
                            "doy",
                            "datetime",
                        ]
                    )[columns]
                    .mean()
                    .reset_index()
                )
                df_calendar_region.loc[:, "average_temperature"] = (
                    df_calendar_region["cpc_tmax"] + df_calendar_region["cpc_tmin"]
                ) / 2.0
                df_calendar_region.set_index(
                    pd.DatetimeIndex(df_calendar_region["datetime"]),
                    inplace=True,
                    drop=True,
                )
                df_calendar_region.index.name = None
                df_calendar_region.sort_values(by="datetime", inplace=True)

                # TODO: Only plot if dates_calendar has valid dates
                # Create title for plot
                sup_title = create_title_for_plot(obj)

                if not df_calendar_region.empty:
                    plot.plots_ts_cur_yr(
                        df_calendar_region,
                        obj.eo_plot,
                        closest=obj.closest,
                        dates_cal=dates_calendar,
                        frcast_yr=obj.plot_season,
                        logos=[obj.logo_harvest, obj.logo_geoglam],
                        dir_out=obj.dir_agmet / "district",
                        sup_title=sup_title,
                        fname=f"{obj.calendar_region}.png",
                    )


def run():
    loop_agmet(
        [
            "D:/Users/ritvik/projects/geoprepare/geoprepare/geoprepare.txt",
            "D:/Users/ritvik/projects/geoprepare/geoprepare/geoextract.txt",
            "D:/Users/ritvik/projects/geocif/geocif/config/geocif.txt",
        ]
    )


if __name__ == "__main__":
    run()
